! ************************************************************************
! Filename : NL_2SMS_TShock_AIR_40Qpvgm.f90
! 
! Author : Timothy Stewart Hills
! Editor : Paul Yoo
! 
! Date Created : August 25, 2016
! Date Edited  : November 16, 2016
! 
! Description : This program will employ Time Path Iteration a la 
! Christian Fisher in order to solve a simple NK model with an interest
! rate rule. There are two two-state Markov shock processes: preference
! shock and government stimulus. The main objective will be to compare the
! IRFs of the variables when there is only a process for the preference
! shock and then compare the IRFs and government multipliers when
! government stimulus is added. We augment the model with a perfectly correlated
! technology shock.
! ************************************************************************

! ************************************************************************
! Command Lines (original peak machines; available until Dec 2016):
! cd /mq/home/m1pcy00/RA_Work/Taisuke_Nakata/ZeroLowerBound/FiscalMultiplier/Fiscal_Multipliers_PostFED/NL_2SMS_TShock/CalibrationExercise/7_2/cPHI_160/HighGShock_8/AltInertialRule
! sh
! . /opt/imsl/frb/bin/fnlsetup.sh
! $FC $F90FLAGS -o NL_2SMS_TShock_AIR_40Qpvgm NL_2SMS_TShock_AIR_40Qpvgm.f90 $LINK_FNL -heap-arrays
! ./NL_2SMS_TShock_AIR_40Qpvgm
! ************************************************************************

! ************************************************************************
! Command Lines (Ubuntu peak machines; available from Dec 2016):
! cd /mq/home/m1pcy00/RA_Work/Taisuke_Nakata/ZeroLowerBound/FiscalMultiplier/Fiscal_Multipliers_PostFED/NL_2SMS_TShock/CalibrationExercise/7_2/cPHI_160/HighGShock_8/AltInertialRule
! sh
! . /etc/profile.d/modules.sh
! module load comp/intel/14.0.5 lib/mkl/11.1.4 lib/imsl/7.1.0
! $FC $F90FLAGS -o NL_2SMS_TShock_AIR_40Qpvgm NL_2SMS_TShock_AIR_40Qpvgm.f90 $LINK_FNL -heap-arrays
! ./NL_2SMS_TShock_AIR_40Qpvgm
! ************************************************************************

! ************************************************************************
! ------------------------------------------------------------------------
! ------------------------------------------------------------------------
! module : params_ss_grid
! 
! description : This module will define the model parameter values, the
! steady state values of the variables, and set up the grids to create a
! discretized state space and for Gaussian-Hermite Quadrature
! ------------------------------------------------------------------------
! ------------------------------------------------------------------------

module params_ss_grid

use Numerical_Libraries

implicit none

! Define model parameters Braun et al and Calibration Exercise values and Tai's original suggestion
double precision, parameter :: cBET     = 0.9925d0 ! 0.997d0 ! 0.9925d0 ! 0.9925d0
double precision, parameter :: cCHIc    = 1d0      ! 1d0     ! 1d0      ! 1d0
double precision, parameter :: cCHIn    = 0.2d0    ! 0.28d0  ! 0.2d0    ! 2d0
double precision, parameter :: cTHETA   = 5d0      ! 7.67d0  ! 5d0      ! 6d0
double precision, parameter :: cPHI     = 160d0    ! 458.4d0 ! 150d0    ! 250d0
double precision, parameter :: cPHIpi   = 3d0      ! 3.46d0  ! 3d0      ! 3d0 
double precision, parameter :: cPHIy    = 1d0      ! 1.63d0  ! 1d0      ! 0d0
double precision, parameter :: cGAM     = 0.2d0

! Steady-state Values
double precision, parameter :: cRBAR  = 1d0 / cBET
double precision, parameter :: cCBAR  = ((cTHETA - 1d0)*(1d0 - cGAM)**(cCHIn)/cTHETA)**(1d0/(cCHIn+cCHIc))
double precision, parameter :: cYBAR  = cCBAR/(1d0-cGAM)
double precision, parameter :: cGBAR  = cGAM*cCBAR/(1d0-cGAM)
double precision, parameter :: cPIBAR = 1d0
double precision, parameter :: cWBAR  = cYBAR**(cCHIn)*cCBAR**(cCHIc)

! Tolerance level for convergence and max iterations
double precision, parameter :: tol  = 1d-10
double precision, parameter :: tol2 = 1d-05
double precision, parameter :: tol3 = 1d-03
integer, parameter          :: max_iter = 50000
integer                     :: i_iter_time, i_converge

! Three variations of the combinations of (high and low) demand, gov't spending, and technology shock magnitudes
integer, parameter  :: n_vars = 3
integer             :: i_vars

! Coefficients for intertial TR rule
double precision                :: cRHOr
integer, parameter              :: n_cRHOr = 4 !100
integer                         :: i_cRHOr
double precision, parameter     :: grid_cRHOr(n_cRHOr) = (/ 0d0, 0.75d0, 0.8d0, 0.85d0 /)
double precision, parameter     :: min_cRHOr = 0d0    !0d0 
double precision, parameter     :: max_cRHOr = 0.85d0 !0.99d0
double precision, parameter     :: step_cRHOr = (max_cRHOr - min_cRHOr) / (n_cRHOr - 1d0)

! Define the two-state Markov process parameters
double precision            :: pLr   ! 0.857 ! 0.902d0 ! Persistence of L shock
!double precision, parameter :: pLr  = 0.75d0 ! 0.857 ! 0.902d0 ! Persistence of L shock
integer, parameter          :: n_pLr = 16
integer                     :: i_pLr
double precision            :: grid_pLr(n_pLr)
double precision, parameter :: min_pLr = 0.825d0
double precision, parameter :: max_pLr = 0.9d0
double precision, parameter :: step_pLr = (max_pLr - min_pLr) / (n_pLr - 1d0)

double precision, parameter :: pHr  = 0.0d0 ! Frequency of L shock

! Define the magnitudes of two-state shocks (demand, gov't spending, and technology)
double precision, parameter :: Hr = 1d0
double precision            :: Lr
double precision            :: crL

double precision, parameter :: cEPS(2) = (/ cGBAR, 1.08d0*cGBAR /)
double precision            :: Hg
double precision, parameter :: Lg = cGBAR

double precision, parameter :: Hz = 1d0
double precision            :: Lz
double precision            :: cLz

double precision, parameter :: cALPHAZ = 0d0
integer, parameter          :: n_cLZ = 2
integer                     :: i_cLZ
double precision            :: grid_cLz(n_cLZ)
double precision, parameter :: min_cLz = 0.927d0
double precision, parameter :: max_cLZ = 1d0
double precision, parameter :: step_cLZ = (max_cLz - min_cLZ)/(n_cLZ - 1d0)

! The parameters that correspond to the two different states
! NOTE: i_Rshock will be used for any state space iterations and j_Rshock
! will be used for computing the conditional expectations
integer, parameter :: n_States = 2
integer, parameter :: n_Rshock = 2
integer, parameter :: n_Gshock = 2
integer, parameter :: n_Zshock = 2

integer :: i_States, j_States
integer :: i_Rshock, j_Rshock

double precision, parameter :: cond_prob_Hr(n_Rshock) = (/ 1d0 - pHr, pHr /)
double precision            :: cond_prob_Lr(n_Rshock)
!double precision, parameter :: cond_prob_Lr(n_Rshock) = (/ 1d0 - pLr, pLr /)
double precision            :: cond_probR_Today(n_Rshock)

double precision :: grid_delshock(n_Rshock)
!double precision, parameter :: grid_delshock(n_Rshock) = (/ Hr, Lr /)

double precision :: grid_Zshock(n_Zshock)
!double precision, parameter :: grid_Zshock(n_Zshock)  = (/ Hz, Lz /)

double precision :: grid_Gshock(n_Gshock)
!double precision, parameter :: grid_Gshock(n_Gshock)  = (/ Hg, Lg /)

! These will define the discretized state space
integer, parameter         :: n_rACTUAL = 51
integer                    :: i_rACTUAL
double precision           :: grid_rACTUAL(n_rACTUAL)
double precision           :: min_rACTUAL
double precision           :: max_rACTUAL
double precision           :: step_rACTUAL

! Trying to find a proper grid interval for rACTUAL
double precision           :: min_rACTUAL_take, max_rACTUAL_take
integer                    :: n_max_min_rACTUAL
integer                    :: i_max_min_rACTUAL

! Total grid points to solve for
integer            :: i_grid_state
integer, parameter :: n_grid_state = n_rACTUAL*n_States

! Global variables: Endogenous state variables
double precision :: rYesterday
double precision :: rACTUALYesterday

! Global variables: Exogenous variables
double precision :: delToday
double precision :: ZToday
double precision :: GToday

! Global variables: Endogenous variables
double precision :: yToday
double precision :: wToday
double precision :: cToday
double precision :: piToday
double precision :: rToday
double precision :: rACTUALToday

double precision :: yTomorrow
double precision :: wTomorrow
double precision :: cTomorrow
double precision :: piTomorrow
double precision :: rTomorrow
double precision :: rACTUALTomorrow

! Conditional Expectations
double precision :: Econs
double precision :: Ephc

double precision :: Econs_int
double precision :: Ephc_int

! Policy functions to be populated
double precision :: pf_y(n_rACTUAL,n_States)
double precision :: pf_c(n_rACTUAL,n_States)
double precision :: pf_w(n_rACTUAL,n_States)
double precision :: pf_pi(n_rACTUAL,n_States)
double precision :: pf_r(n_rACTUAL,n_States)
double precision :: pf_rACTUAL(n_rACTUAL,n_States)

!$OMP THREADPRIVATE( i_grid_state, i_rACTUAL, i_States, j_States, i_Rshock, j_Rshock, cond_probR_Today, rYesterday, rACTUALYesterday, delToday, GToday, ZToday, yToday, yTomorrow, wToday, wTomorrow, cToday, cTomorrow, piToday, piTomorrow, rToday, rTomorrow, rACTUALToday, rACTUALTomorrow, Econs, Econs_int, Ephc, Ephc_int )

end module params_ss_grid

! ************************************************************************
! ------------------------------------------------------------------------
! ------------------------------------------------------------------------
! program : INERTIAL
! 
! description : This program will use time path iteration a la Christiano 
! Fisher approach to solve the different model specifications
! ------------------------------------------------------------------------
! ------------------------------------------------------------------------

program INERTIAL

use Numerical_Libraries
use params_ss_grid
use omp_lib

implicit none

integer :: i_stat

! prep for nonlinear solver
integer, parameter  :: n_eq = 2
integer, parameter  :: n_eq_ZLB = 2
double precision    :: tol_eqm = 1d-32
integer, parameter  :: max_iter_eqm = 10000
double precision    :: x_guess(n_eq),x_out(n_eq),fnorm

! Policy functions to be populated
double precision :: pf_y_up(n_rACTUAL,n_States)
double precision :: pf_c_up(n_rACTUAL,n_States)
double precision :: pf_w_up(n_rACTUAL,n_States)
double precision :: pf_pi_up(n_rACTUAL,n_States)
double precision :: pf_r_up(n_rACTUAL,n_States)
double precision :: pf_rACTUAL_up(n_rACTUAL,n_States)

! Prep to match consumption decline
integer                     :: i_decline_check
double precision            :: shock_adjust
double precision, parameter :: Standard_Decline = -0.07d0
double precision            :: decline
double precision            :: current_crL

! Prep to match inflation decline
integer                     :: i_inf_decline_check
double precision            :: z_shock_adjust
double precision, parameter :: Inf_Standard_Decline = -2d0
double precision            :: inf_decline

! Global Variables
double precision :: diff_y
double precision :: diff_c
double precision :: diff_w
double precision :: diff_pi
double precision :: diff_r
double precision :: diff_rACTUAL

double precision :: max_diff

! Time the computation
integer :: beginning, end, rate

external linear_interp_1S
external eqm_solve
external initial_declines
external irf_delshock
external irf_gshock
external irf_delshock_gshock

call omp_set_num_threads(7)

call system_clock(beginning,rate)

! do i_cRHOr = 1,n_cRHOr
!     grid_cRHOr(i_cRHOr) = min_cRHOr + (dble(i_cRHOr)-1d0)*step_cRHOr
! end do

do i_pLr = 1,n_pLr
    grid_pLr(i_pLr) = min_pLr + (dble(i_pLr) - 1d0)*step_pLr
end do

do i_cLz = 1,n_cLz
    grid_cLz(i_cLz) = min_cLz + (dble(i_cLz) - 1d0)*step_cLz
end do

current_crL = 1.01936620715051d0 ! 1.02117837890494d0
cLz         = 0.939198208260574d0   ! 0.926792238479705d0 ! 0.96d0

! ******************************************************************************
! ************************** Start of cRHOr loop *******************************
! ******************************************************************************

do i_cRHOr = 1,n_cRHOr ! 1,n_cRHOr

		if (i_cRHOr == 1) then
        !open(unit = 1, file = 'irf_G_MODEL_delshock_NITR_corr_FM_NL_AIR_40Qpvgm.dat', status = 'replace', action = 'write', iostat = i_stat)
        !open(unit = 2, file = 'irf_G_MODEL_gshock_NITR_corr_FM_NL_AIR_40Qpvgm.dat', status = 'replace', action = 'write', iostat = i_stat)
        !open(unit = 3, file = 'irf_G_MODEL_delshock_gshock_NITR_corr_FM_NL_AIR_40Qpvgm.dat', status = 'replace', action = 'write', iostat = i_stat)
        open(unit = 4, file = 'irf_G_MODEL_delshock_NITR_e_corr_FM_NL_AIR_40Qpvgm.dat', status = 'replace', action = 'write', iostat = i_stat)
        open(unit = 5, file = 'irf_G_MODEL_gshock_NITR_e_corr_FM_NL_AIR_40Qpvgm.dat', status = 'replace', action = 'write', iostat = i_stat)
        open(unit = 6, file = 'irf_G_MODEL_delshock_gshock_NITR_e_corr_FM_NL_AIR_40Qpvgm.dat', status = 'replace', action = 'write', iostat = i_stat)    
        current_crL = 1.01936620715051d0
        cLz         = 0.939198208260574d0   
		else if (i_cRHOr == 2) then
        !open(unit = 1, file = 'irf_G_MODEL_delshock_NITR_corr_FM_NL_AIR_40Qpvgm.dat', position="append", action = 'write', iostat = i_stat)
        !open(unit = 2, file = 'irf_G_MODEL_gshock_NITR_corr_FM_NL_AIR_40Qpvgm.dat', position="append", action = 'write', iostat = i_stat)
        !open(unit = 3, file = 'irf_G_MODEL_delshock_gshock_NITR_corr_FM_NL_AIR_40Qpvgm.dat', position="append", action = 'write', iostat = i_stat)
        open(unit = 4, file = 'irf_G_MODEL_delshock_NITR_e_corr_FM_NL_AIR_40Qpvgm.dat', position="append", action = 'write', iostat = i_stat)
        open(unit = 5, file = 'irf_G_MODEL_gshock_NITR_e_corr_FM_NL_AIR_40Qpvgm.dat', position="append", action = 'write', iostat = i_stat)
        open(unit = 6, file = 'irf_G_MODEL_delshock_gshock_NITR_e_corr_FM_NL_AIR_40Qpvgm.dat', position="append", action = 'write', iostat = i_stat)
        current_crL = 1.02071963066586d0
        cLz         = 0.941590931472106d0
		end if

    ! ******************************************************************************
    ! ************************** Start of pLr loop *********************************
    ! ******************************************************************************
    		    
    do i_pLr = 1,1
    
        pLr = grid_pLr(i_pLr)
    
        write(*,*) "^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^"
        write(*,*) "Persistence = ", pLr
        write(*,*) "^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^"
        write(*,*) " "
    
        cond_prob_Lr(:) = (/ 1d0 - pLr, pLr /)

        write(*,*) "\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\"
        write(*,*) "Solving for the SHADOW (Inertial) Taylor Rule"
        write(*,*) "///////////////////////////////////////////////"
        write(*,*) " "
        cRHOr       = grid_cRHOr( i_cRHOr ) 
        write(*,*) "------------------------------------------------------"
        write(*,*) "cRHOr = ",cRHOr
        write(*,*) "------------------------------------------------------"

        if (cRHOr .EQ. 0d0) then 
            n_max_min_rACTUAL = 1
        else
            n_max_min_rACTUAL = 2
        end if

        ! ******************************************************************************
        ! ****************** Start of finding better grid for rACTUAL ******************
        ! ******************************************************************************

        do i_max_min_rACTUAL = 1,n_max_min_rACTUAL
            
            crL = current_crL

            ! Set the initial guesses
            write(*,*) "Setting the initial PF guesses"
            write(*,*) " "
            do i_rACTUAL = 1,n_rACTUAL
                do i_States = 1,n_States
                    pf_y(i_rACTUAL,i_States)       = cYBAR
                    pf_c(i_rACTUAL,i_States)       = cCBAR
                    pf_w(i_rACTUAL,i_States)       = cWBAR
                    pf_pi(i_rACTUAL,i_States)      = cPIBAR
                    pf_r(i_rACTUAL,i_States)       = cRBAR
                    pf_rACTUAL(i_rACTUAL,i_States) = cRBAR
                end do
            end do

            ! ************************************************************************************************************************************
            ! ****** Start of three variations of the combinations of (high and low) demand, gov't spending, and technology shock magnitudes *****
            ! ************************************************************************************************************************************

            do i_vars = 1,n_vars

                if (i_max_min_rACTUAL .EQ. 1) then                    
                    ! The initial grid for rACTUAL, which will be improved when i_max_min_rACTUAL = 2
                    min_rACTUAL = 1d0             ! cRBAR           ! 0.965605d0
                    max_rACTUAL = cRBAR + 0.001d0 ! cRBAR + 0.001d0
                    step_rACTUAL = (max_rACTUAL - min_rACTUAL) / (n_rACTUAL - 1d0)
                    do i_rACTUAL = 1,n_rACTUAL
                        grid_rACTUAL(i_rACTUAL) = min_rACTUAL + (dble(i_rACTUAL)-1d0)*step_rACTUAL
                    end do
                else                	
                    if (i_vars < 3) then                        
                        ! For the cases where low state demand shock is set to be contractionary
                        
                        ! Discretize the State Space
                        write(*,*) "Discretizing the state space"
                        write(*,*) " "

                        write(*,*) "SHADOW INTEREST RATE GRID"
                        write(*,*) " "
                        min_rACTUAL = MIN(min_rACTUAL_take,cRbar)
                        max_rACTUAL = MAX(min_rACTUAL_take,cRbar)
                        step_rACTUAL = (max_rACTUAL - min_rACTUAL) / (n_rACTUAL - 1d0)
                        do i_rACTUAL = 1,n_rACTUAL
                            grid_rACTUAL(i_rACTUAL) = min_rACTUAL + (dble(i_rACTUAL)-1d0)*step_rACTUAL
                        end do
                    else                        
                        ! For the case where low state demand shock is not contractionary but take its the steady state value
                        
                        ! Discretize the State Space
                        write(*,*) "Discretizing the state space"
                        write(*,*) " "

                        write(*,*) "SHADOW INTEREST RATE GRID"
                        write(*,*) " "
                        min_rACTUAL = MIN(cRbar,max_rACTUAL_take)
                        max_rACTUAL = MAX(cRbar,max_rACTUAL_take)
                        step_rACTUAL = (max_rACTUAL - min_rACTUAL) / (n_rACTUAL - 1d0)
                        do i_rACTUAL = 1,n_rACTUAL
                            grid_rACTUAL(i_rACTUAL) = min_rACTUAL + (dble(i_rACTUAL)-1d0)*step_rACTUAL
                        end do

                        do i_rACTUAL = 1,n_rACTUAL
                            do i_States = 1,n_States
                                pf_y(i_rACTUAL,i_States)       = cYBAR
                                pf_c(i_rACTUAL,i_States)       = cCBAR
                                pf_w(i_rACTUAL,i_States)       = cWBAR
                                pf_pi(i_rACTUAL,i_States)      = cPIBAR
                                pf_r(i_rACTUAL,i_States)       = cRBAR
                                pf_rACTUAL(i_rACTUAL,i_States) = cRBAR
                            end do
                        end do
                    end if 
                end if

                ! ******************************************************************************************************************************
                ! ************ Start of matching the initial decline in inflation by varying low-state Technology shock magnitude **************
                ! ******************************************************************************************************************************
                
                i_inf_decline_check = 0
    
                do while (i_inf_decline_check .EQ. 0)
    
                    write(*,*) "^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^"
                    write(*,*) "Technology Shock = ", cLz
                    write(*,*) "^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^"
                    write(*,*) " "

                    ! ******************************************************************************************************************************
                    ! ************* Start of matching the initial decline in consumption by varying low-state demand shock magnitude ***************
                    ! ******************************************************************************************************************************
                    
                    i_decline_check = 0

                    do while ( i_decline_check .EQ. 0 ) 

                        ! ******************************************************************************
                        ! ****************** Start of solving for the policy functions *****************
                        ! ******************************************************************************
        
                        i_converge  = 0
                        i_iter_time = 1

                        do while (i_converge == 0 .and. i_iter_time <= max_iter)

                            if ( i_vars .EQ. 1 ) then
                                Hg = cEPS(1)
                                Lr = crL
                                Lz = cLz
                            else if ( i_vars .EQ. 2) then
                                Hg = cEPS(2)
                                Lr = crL
                                Lz = cLz
                            else
                                Hg = cEPS(2)
                                Lr = 1d0
                                Lz = 1d0
                            end if
                    
                            grid_delshock   = (/ Hr, Lr /)
                            grid_Gshock     = (/ Hg, Lg /)
                            grid_Zshock     = (/ Hz, Lz /)
                            
                            !$OMP PARALLEL DO PRIVATE( tol_eqm, x_guess, x_out, fnorm )
                            do i_grid_state = 1,n_grid_state
            		                ! ----------------------------------
            		                ! Define the today's state variables
            		                ! ----------------------------------
                                i_rACTUAL = MOD(i_grid_state,n_rACTUAL) + 1
                                i_States  = MOD(CEILING((dble(i_grid_state) / dble(n_rACTUAL))),n_States) + 1

                                rACTUALYesterday = grid_rACTUAL(i_rACTUAL)
                                
                                if (i_States == 1) then
                                    ! Normal State (Watch out for the GShock..)
                                    cond_probR_Today = cond_prob_Hr
                                    delToday         = grid_delshock(1)
                                    GToday           = grid_Gshock(2)
                                    ZToday           = grid_Zshock(1)
                                else
                                    ! Crisis State (Watch out for the GShock..)
                                    cond_probR_Today = cond_prob_Lr
                                    delToday         = grid_delshock(2)
                                    GToday           = grid_Gshock(1)
                                    ZToday           = grid_Zshock(2)
                                end if
                                
                                ! ------------------------------
                                ! Solve for the non-binding case
                                ! ------------------------------
                                x_guess(1) = pf_pi(i_rACTUAL,i_States)
                                x_guess(2) = pf_c(i_rACTUAL,i_States)

                                call erset(0,0,0)
                                call dneqnf(eqm_solve,tol_eqm,n_eq,max_iter_eqm,x_guess,x_out,fnorm)

                                piToday = x_out(1)
                                cToday  = x_out(2)

                                yToday       = (cToday + GToday)/(1d0-cPHI/2d0*(piToday-1d0)**2d0)
                                wToday       = (yToday/ZToday)**(cCHIn)*cToday**(cCHIc)
                                rToday       = MAX(1d0,cRBAR*(rACTUALYesterday/cRBAR)**(cRHOr)*((piToday/cPIBAR)**(cPHIpi)*(yToday/cYBAR)**(cPHIy))**(1d0 - cRHOr))
                                rACTUALToday = rToday

                                ! -----------------------------------
                                ! Prepare the policy function updates
                                ! -----------------------------------
                                pf_r_up(i_rACTUAL,i_States)       = rToday
                                pf_rACTUAL_up(i_rACTUAL,i_States) = rACTUALToday
                                pf_pi_up(i_rACTUAL,i_States)      = piToday
                                pf_y_up(i_rACTUAL,i_States)       = yToday
                                pf_c_up(i_rACTUAL,i_States)       = cToday
                                pf_w_up(i_rACTUAL,i_States)       = wToday

                            end do
                            !$OMP END PARALLEL DO

                            ! Find the difference between the policy functions and the updates
                            diff_r       = sum(real(abs(pf_r - pf_r_up)))
                            diff_c       = sum(real(abs(pf_c - pf_c_up)))
                            diff_w       = sum(real(abs(pf_w - pf_w_up)))
                            diff_rACTUAL = sum(real(abs(pf_rACTUAL - pf_rACTUAL_up)))
                            diff_y       = sum(real(abs(pf_y - pf_y_up)))
                            diff_pi      = sum(real(abs(pf_pi - pf_pi_up)))

                            ! Check for convergence
                            max_diff = diff_r + diff_c + diff_w + diff_rACTUAL + diff_y + diff_pi

                            i_iter_time = i_iter_time + 1

                            if (max_diff < tol) then
                                i_converge = 1
                                write(*,*) "------------------------------------------------------"
                                write(*,*) "AT ITERATION = ", i_iter_time-1
                                write(*,*) "MAX DIFFERENCE = ", max_diff
                                write(*,*) "------------------------------------------------------"
                                write(*,*) " "
                            end if

                            ! ---------------------------
                            ! Update the policy functions
                            ! ---------------------------
                            pf_r        = pf_r_up
                            pf_rACTUAL  = pf_rACTUAL_up
                            pf_y        = pf_y_up
                            pf_c        = pf_c_up
                            pf_w        = pf_w_up
                            pf_pi       = pf_pi_up

                            ! if (mod(i_iter_time-1,1000) == 0) then
                            !     write(*,*) "------------------------------------------------------"
                            !     write(*,*) "AT ITERATION   = ", i_iter_time-1
                            !     write(*,*) "MAX DIFFERENCE = ", max_diff
                            !     write(*,*) "DFFERENCE COMPONENTS: ",diff_r,diff_rACTUAL,diff_Z,diff_y,diff_pi
                            !     write(*,*) "------------------------------------------------------"
                            !     write(*,*) " "
                            ! end if

                        end do

                        ! ******************************************************************************
                        ! ****************** End of solving for the policy functions *******************
                        ! ******************************************************************************

                        ! Initial Declines
                        call initial_declines(decline,inf_decline)
                        
                        ! Compute the required shock adjustments
                        shock_adjust   = ((decline - Standard_Decline)/Standard_Decline)
                        z_shock_adjust = ((inf_decline - inf_standard_decline)/inf_standard_decline)
                        
                        ! Initial Decline in output (7%)
                        if (ABS(decline - Standard_Decline) < tol2) then
                            i_decline_check = 1
                        else if (i_vars .NE. 1) then
                            ! We set crL that gives 7% and 1% initial declines only to the case when gov't spending is the same across two states.
                            i_decline_check = 1
                        else
                            crL = crL - shock_adjust*ABS((crL-1d0)/1d0)*0.01d0
                        end if
                        
                        write(*,*) "Decline = ", decline
                        write(*,*) "Updated Preference Shock = ", crL

                    end do

                    ! ******************************************************************************************************************************
                    ! ************* End of matching the initial decline in consumption by varying low-state demand shock magnitude *****************
                    ! ******************************************************************************************************************************
                    
                    ! check inflation decline at 1%
                    if ( ABS(inf_decline - inf_Standard_Decline) < tol3 ) then
                        i_inf_decline_check = 1
                    else if ( i_vars .NE. 1 ) then
                        ! We set crL that gives 7% and 1% initial declines only to the case when gov't spending is the same across two states.
                        i_inf_decline_check = 1
                    else
                        cLz = cLz - z_shock_adjust*0.01d0
                    end if
                    current_crL = crL
                    write(*,*) "Inf Decline =", inf_decline, " Updated Z Shock = ", cLz
                    write(*,*) " "

                end do ! i_inf_decline_check

                ! ******************************************************************************************************************************
                ! ************ End of matching the initial decline in inflation by varying low-state Technology shock magnitude ****************
                ! ******************************************************************************************************************************
                
                ! Compute the IRFs
                write(*,*) "Computing the IRFs"
                write(*,*) " "
                if (i_vars .EQ. 1) then
                    call irf_delshock()
                else if (i_vars .EQ. 2) then
                    call irf_delshock_gshock()
                else
                    call irf_gshock()
                end if

            end do ! i_vars

            ! ************************************************************************************************************************************
            ! ****** End of three variations of the combinations of (high and low) demand, gov't spending, and technology shock magnitudes *******
            ! ************************************************************************************************************************************

        end do ! i_max_min_rACTUAL

        ! ******************************************************************************
        ! ****************** End of finding better grid for rACTUAL ********************
        ! ******************************************************************************
            
    end do ! i_pLr

    ! ******************************************************************************
    ! ************************** End of pLr loop ***********************************
    ! ******************************************************************************
    
end do ! i_cRHOr

! ******************************************************************************
! ************************** End of cRHOr loop *********************************
! ******************************************************************************

call system_clock(end)

write(*,*) "**************************************************************"
write(*,*) "Total elapsed time for program = ", real(end - beginning) / real(rate)," seconds"
write(*,*) "**************************************************************"
write(*,*) " "

!close(1)
!close(2)
!close(3)
close(4)
close(5)
close(6)

write(*,*) "----------------------------"
write(*,*) "*******END OF PROGRAM*******"
write(*,*) "----------------------------"
write(*,*) " "

end program INERTIAL

! ************************************************************************
! ------------------------------------------------------------------------
! ------------------------------------------------------------------------
! subroutine : eqm_solve
! 
! description : solves the non-linear system of equations conditional on 
! the nonbinding case.
! ------------------------------------------------------------------------
! ------------------------------------------------------------------------

subroutine eqm_solve(x_guess,residuals,n_eq)

use params_ss_grid
use Numerical_Libraries

implicit none

integer, intent(in)             :: n_eq
double precision, intent(in)    :: x_guess(n_eq)
double precision, intent(out)   :: residuals(n_eq)

external linear_interp_1S

piToday = x_guess(1)
cToday  = x_guess(2)

yToday       = (cToday + GToday)/(1d0-cPHI/2d0*(piToday-1d0)**2d0)
wToday       = (yToday/ZToday)**(cCHIn)*cToday**(cCHIc)
rToday       = MAX(1d0,cRBAR*(rACTUALYesterday/cRBAR)**(cRHOr)*((piToday/cPIBAR)**(cPHIpi)*(yToday/cYBAR)**(cPHIy))**(1d0 - cRHOr))
rACTUALToday = rToday

Econs_int = 0d0
Ephc_int = 0d0

! Compute the conditional expectations
do j_States = 1,n_States
    if (j_States == 1) then
    	  ! Normal
        j_Rshock = 1
    else 
    	  ! Crisis
        j_Rshock = 2
    end if

    call linear_interp_1S(grid_rACTUAL,rACTUALToday,n_rACTUAL,pf_pi(:,j_States),piTomorrow)
    call linear_interp_1S(grid_rACTUAL,rACTUALToday,n_rACTUAL,pf_y(:,j_States),yTomorrow)
    call linear_interp_1S(grid_rACTUAL,rACTUALToday,n_rACTUAL,pf_c(:,j_States),cTomorrow)

    Econs   = cTomorrow**(-cCHIc)/piTomorrow
    Ephc    = yTomorrow*cTomorrow**(-cCHIc)*cPHI*(piTomorrow - 1d0)*piTomorrow

    Econs_int   = Econs_int + cond_probR_Today(j_Rshock)*Econs
    Ephc_int    = Ephc_int + cond_probR_Today(j_Rshock)*Ephc    
end do

residuals(1) = (cToday**(-cCHIc) - (cBET*delToday*rToday*Econs_int))
residuals(2) = (yToday*cToday**(-cCHIc)*(cPHI*(piToday - 1d0)*piToday - (1d0 - cTHETA) - cTHETA*wToday/zToday) - (cBET*delToday*Ephc_int))

return

end subroutine eqm_solve

! ************************************************************************
! ------------------------------------------------------------------------
! ------------------------------------------------------------------------
! subroutine : initial_declines
! 
! description : Computes the intial decline in output as measured when
! there is a preference shock.
! ------------------------------------------------------------------------
! ------------------------------------------------------------------------

subroutine initial_declines(decline,inf_decline)

use params_ss_grid

implicit none

double precision, intent(out) :: decline
double precision, intent(out) :: inf_decline

rACTUALYesterday = cRbar

call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_y(:,2),yToday)
call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_pi(:,2),piToday)

decline     = (yToday-cYBAR)/cYBAR
inf_decline = 400d0*LOG(piToday)

return

end subroutine initial_declines

! ************************************************************************
! ------------------------------------------------------------------------
! ------------------------------------------------------------------------
! subroutine : irf_delshock
! 
! description : computes the impulse response functions of the two-state
! MARKOV shock process assuming TAYLOR Equilibrium.
! ------------------------------------------------------------------------
! ------------------------------------------------------------------------

subroutine irf_delshock()

use params_ss_grid

implicit none

integer, parameter :: per       = 40
integer, parameter :: n_shock   = 1
integer            :: years(per)
integer            :: i_year
integer            :: i_sim

double precision :: sim_y(per)
double precision :: sim_c(per)
double precision :: sim_w(per)
double precision :: sim_pi(per)
double precision :: sim_S(per,2)
double precision :: sim_r(per)
double precision :: sim_rACTUAL(per)

double precision :: sim_E1dYT(per)
double precision :: sim_E1dGT(per)
double precision :: sim_E1IT(per)
double precision :: sim_E1PT(per)
double precision :: sim_E1RT(per)

integer, parameter :: max_e = per

double precision            :: diff_rsm
double precision, parameter :: tol_diff = 1d-10

double precision :: E1IT, E1PT, E1dYT, E1dGT

external E1PT
external E1IT
external E1dYT
external E1dGT

do i_year = 1,per
    years(i_year) = i_year
end do

diff_rsm = 1d0
rACTUALYesterday = cRbar

if ((i_max_min_rACTUAL .EQ. 1).AND.(cRHOr .NE. 0d0)) then
    ! Trying to find the minimum of the grid for rACTUALYesterday
    do while (ABS(diff_rsm)>tol_diff)        
        call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_rACTUAL(:,2),rACTUALToday)
        diff_rsm = rACTUALToday - rACTUALYesterday
        rACTUALYesterday = rACTUALToday    
    end do
    min_rACTUAL_take = rACTUALToday-tol_diff
else
    do i_sim = 1,n_shock
        ! sim_S(i_sim,:) = (/ Lr, Hg /)
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_y(:,2),sim_y(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_c(:,2),sim_c(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_w(:,2),sim_w(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_pi(:,2),sim_pi(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_r(:,2),sim_r(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_rACTUAL(:,2),sim_rACTUAL(i_sim))
        sim_E1dYT(i_sim) = E1dYT(2,i_sim,cRbar)
        sim_E1dGT(i_sim) = E1dGT(2,i_sim,cRbar)
        ! sim_E1IT(i_sim)  = E1IT(2,i_sim,cRbar)
        ! sim_E1PT(i_sim)  = E1PT(2,i_sim,cRbar)
        ! sim_E1RT(i_sim)  = E1IT(2,i_sim,cRbar) - E1PT(2,i_sim+1,cRbar)
        ! rACTUALYesterday = sim_rACTUAL(i_sim)
    end do

    do i_sim = n_shock+1,per
        sim_S(i_sim,:) = (/ Hr, Lg /)
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_y(:,1),sim_y(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_c(:,1),sim_c(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_w(:,1),sim_w(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_pi(:,1),sim_pi(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_r(:,1),sim_r(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_rACTUAL(:,1),sim_rACTUAL(i_sim))
        if (i_sim .LE. max_e) then
            sim_E1dYT(i_sim) = E1dYT(2,i_sim,cRbar)
            sim_E1dGT(i_sim) = E1dGT(2,i_sim,cRbar)
            ! sim_E1IT(i_sim) = E1IT(2,i_sim,cRbar)
            ! sim_E1PT(i_sim) = E1PT(2,i_sim,cRbar)
            ! sim_E1RT(i_sim) = E1IT(2,i_sim,cRbar) - E1PT(2,i_sim+1,cRbar)
        end if
        ! rACTUALYesterday = sim_rACTUAL(i_sim)
    end do

    100 format(i3,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15)
    400 format(f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15)

    write(*,*) " "
    write(*,*) "----------------------------------------------------------"
    write(*,*) "Writing the irf_MPs to a _NITR_corr_40Qpvgm.dat file written the following way: "
    write(*,*) "Period, cRHOr,  S, y, pi, i, iSHADOW, Z"
    write(*,*) "----------------------------------------------------------"

    ! do i_sim = 1,per
    !     write(1,100) years(i_sim), cLz, 100d0*(sim_S(i_sim,1)-1d0)/1d0,100d0*(sim_S(i_sim,2)-cGBAR)/cGBAR, 100d0*(sim_y(i_sim)-cYBAR)/cYBAR, 100d0*(sim_c(i_sim)-cCBAR)/cCBAR, 400*(LOG(sim_pi(i_sim))), 400*LOG(sim_r(i_sim)), 400*LOG(sim_rACTUAL(i_sim)),100d0*(sim_w(i_sim)-cWBAR)/cWBAR, cRHOr
    ! end do
    
    write(4,400) cRHOr, SUM(sim_E1dYT(:)), SUM(sim_E1dGT(:))

end if

end subroutine irf_delshock

! ************************************************************************
! ------------------------------------------------------------------------
! ------------------------------------------------------------------------
! subroutine : irf_gshock
! 
! description : computes the impulse response functions of the two-state
! MARKOV shock process assuming TAYLOR Equilibrium.
! ------------------------------------------------------------------------
! ------------------------------------------------------------------------

subroutine irf_gshock()

use params_ss_grid

implicit none

integer, parameter :: per       = 40
integer, parameter :: n_shock   = 1
integer            :: years(per)
integer            :: i_year
integer            :: i_sim

double precision :: sim_y(per)
double precision :: sim_c(per)
double precision :: sim_w(per)
double precision :: sim_pi(per)
double precision :: sim_S(per,2)
double precision :: sim_r(per)
double precision :: sim_rACTUAL(per)

double precision :: sim_E1dYT(per)
double precision :: sim_E1dGT(per)
double precision :: sim_E1IT(per)
double precision :: sim_E1PT(per)
double precision :: sim_E1RT(per)

integer, parameter :: max_e = per

double precision            :: diff_rsm
double precision, parameter :: tol_diff = 1d-10

double precision :: E1IT, E1PT, E1dYT, E1dGT

external E1PT
external E1IT
external E1dYT
external E1dGT

do i_year = 1,per
    years(i_year) = i_year
end do

diff_rsm = 1d0
rACTUALYesterday = cRbar

if ((i_max_min_rACTUAL .EQ. 1).AND.(cRHOr .NE. 0d0)) then
    ! Trying to find the maximum of the grid for rACTUALYesterday
    do while (ABS(diff_rsm)>tol_diff)        
        call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_rACTUAL(:,2),rACTUALToday)
        diff_rsm = rACTUALToday - rACTUALYesterday
        rACTUALYesterday = rACTUALToday
    end do
    max_rACTUAL_take = rACTUALToday+tol_diff
else
    do i_sim = 1,n_shock
        ! sim_S(i_sim,:) = (/ Lr, Hg /)
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_y(:,2),sim_y(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_c(:,2),sim_c(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_w(:,2),sim_w(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_pi(:,2),sim_pi(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_r(:,2),sim_r(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_rACTUAL(:,2),sim_rACTUAL(i_sim))
        sim_E1dYT(i_sim) = E1dYT(2,i_sim,cRbar)
        sim_E1dGT(i_sim) = E1dGT(2,i_sim,cRbar)
        ! sim_E1IT(i_sim)  = E1IT(2,i_sim,cRbar)
        ! sim_E1PT(i_sim)  = E1PT(2,i_sim,cRbar)
        ! sim_E1RT(i_sim)  = E1IT(2,i_sim,cRbar) - E1PT(2,i_sim+1,cRbar)
        ! rACTUALYesterday = sim_rACTUAL(i_sim)
    end do

    do i_sim = n_shock+1,per
        ! sim_S(i_sim,:) = (/ Hr, Lg /)
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_y(:,1),sim_y(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_c(:,1),sim_c(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_w(:,1),sim_w(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_pi(:,1),sim_pi(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_r(:,1),sim_r(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_rACTUAL(:,1),sim_rACTUAL(i_sim))
        if (i_sim .LE. max_e) then
            sim_E1dYT(i_sim) = E1dYT(2,i_sim,cRbar)
            sim_E1dGT(i_sim) = E1dGT(2,i_sim,cRbar)
            ! sim_E1IT(i_sim) = E1IT(2,i_sim,cRbar)
            ! sim_E1PT(i_sim) = E1PT(2,i_sim,cRbar)
            ! sim_E1RT(i_sim) = E1IT(2,i_sim,cRbar) - E1PT(2,i_sim+1,cRbar)
        end if
        ! rACTUALYesterday = sim_rACTUAL(i_sim)
    end do

    200 format(i3,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15)
    500 format(f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15)
    
    write(*,*) " "
    write(*,*) "----------------------------------------------------------"
    write(*,*) "Writing the irf_MPs to a _NITR_corr_40Qpvgm.dat file written the following way: "
    write(*,*) "Period, cRHOr, S, y, pi, i, iSHADOW"
    write(*,*) "----------------------------------------------------------"

    ! do i_sim = 1,per
    !     write(2,200) years(i_sim), cLz,  100d0*(sim_S(i_sim,1)-1d0)/1d0,100d0*(sim_S(i_sim,2)-cGBAR)/cGBAR, 100d0*(sim_y(i_sim)-cYBAR)/cYBAR, 100d0*(sim_c(i_sim)-cCBAR)/cCBAR, 400*log(sim_pi(i_sim)), 400*log(sim_r(i_sim)), 400*log(sim_rACTUAL(i_sim)),100d0*(sim_w(i_sim)-cWBAR)/cWBAR, cRHOr
    ! end do
    
    write(5,500) cRHOr, SUM(sim_E1dYT(:)), SUM(sim_E1dGT(:))

end if

end subroutine irf_gshock

! ************************************************************************
! ------------------------------------------------------------------------
! ------------------------------------------------------------------------
! subroutine : irf_delshock_gshock
! 
! description : computes the impulse response functions of the two-state
! MARKOV shock process assuming TAYLOR Equilibrium.
! ------------------------------------------------------------------------
! ------------------------------------------------------------------------

subroutine irf_delshock_gshock()

use params_ss_grid

implicit none

integer, parameter :: per       = 40
integer, parameter :: n_shock   = 1
integer            :: years(per)
integer            :: i_year
integer            :: i_sim

double precision :: sim_y(per)
double precision :: sim_c(per)
double precision :: sim_w(per)
double precision :: sim_pi(per)
double precision :: sim_S(per,2)
double precision :: sim_r(per)
double precision :: sim_rACTUAL(per)

double precision :: sim_E1dYT(per)
double precision :: sim_E1dGT(per)
double precision :: sim_E1IT(per)
double precision :: sim_E1PT(per)
double precision :: sim_E1RT(per)

integer, parameter :: max_e = per

double precision            :: diff_rsm
double precision, parameter :: tol_diff = 1d-10

double precision :: E1IT, E1PT, E1dYT, E1dGT

external E1PT
external E1IT
external E1dYT
external E1dGT

do i_year = 1,per
    years(i_year) = i_year
end do

rACTUALYesterday = cRbar

if ((i_max_min_rACTUAL .EQ. 1).AND.(cRHOr .NE. 0d0)) then
    ! Already have a decent grid interval for rACTUALYesterday from irf_delshock()
else
    do i_sim = 1,n_shock
        ! sim_S(i_sim,:) = (/ Lr, Hg /)
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_y(:,2),sim_y(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_c(:,2),sim_c(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_w(:,2),sim_w(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_pi(:,2),sim_pi(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_r(:,2),sim_r(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_rACTUAL(:,2),sim_rACTUAL(i_sim))
        sim_E1dYT(i_sim) = E1dYT(2,i_sim,cRbar)
        sim_E1dGT(i_sim) = E1dGT(2,i_sim,cRbar)
        ! sim_E1IT(i_sim)  = E1IT(2,i_sim,cRbar)
        ! sim_E1PT(i_sim)  = E1PT(2,i_sim,cRbar)
        ! sim_E1RT(i_sim)  = E1IT(2,i_sim,cRbar) - E1PT(2,i_sim+1,cRbar)
        ! rACTUALYesterday = sim_rACTUAL(i_sim)
    end do

    do i_sim = n_shock+1,per
        ! sim_S(i_sim,:) = (/ Hr, Lg /)
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_y(:,1),sim_y(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_c(:,1),sim_c(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_w(:,1),sim_w(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_pi(:,1),sim_pi(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_r(:,1),sim_r(i_sim))
        ! call linear_interp_1S(grid_rACTUAL,rACTUALYesterday,n_rACTUAL,pf_rACTUAL(:,1),sim_rACTUAL(i_sim))
        if (i_sim .LE. max_e) then
            sim_E1dYT(i_sim) = E1dYT(2,i_sim,cRbar)
            sim_E1dGT(i_sim) = E1dGT(2,i_sim,cRbar)
            ! sim_E1IT(i_sim) = E1IT(2,i_sim,cRbar)
            ! sim_E1PT(i_sim) = E1PT(2,i_sim,cRbar)
            ! sim_E1RT(i_sim) = E1IT(2,i_sim,cRbar) - E1PT(2,i_sim+1,cRbar)
        end if
        ! rACTUALYesterday = sim_rACTUAL(i_sim)
    end do

    300 format(i3,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15)
    600 format(f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15,2x,f25.15)

    write(*,*) " "
    write(*,*) "----------------------------------------------------------"
    write(*,*) "Writing the irf_MPs to a _NITR_corr_40Qpvgm.dat file written the following way: "
    write(*,*) "Period, cRHOr, cALPHAZ, S, y, pi, i, iSHADOW, Z"
    write(*,*) "----------------------------------------------------------"

    ! do i_sim = 1,per
    !     write(3,300) years(i_sim), cLz,  100d0*(sim_S(i_sim,1)-1d0)/1d0, 100d0*(sim_S(i_sim,2)-cGBAR)/cGBAR, 100d0*(sim_y(i_sim)-cYBAR)/cYBAR, 100d0*(sim_c(i_sim)-cCBAR)/cCBAR, 400*log(sim_pi(i_sim)), 400*log(sim_r(i_sim)), 400*log(sim_rACTUAL(i_sim)) ,100d0*(sim_w(i_sim)-cWBAR)/cWBAR, cRHOr
    ! end do
    
    write(6,600) cRHOr, SUM(sim_E1dYT(:)), SUM(sim_E1dGT(:))

end if

end subroutine irf_delshock_gshock

! ************************************************************************
! ------------------------------------------------------------------------
! ------------------------------------------------------------------------
! subroutine : linear_interp_1S
! 
! description : employs the methods of linear interpolation. Originally
! adapted from MATLAB code by (c) Richter, Throckmorton, and Walker, 2013.
! This version only has one state variable.
! ------------------------------------------------------------------------
! ------------------------------------------------------------------------

subroutine linear_interp_1S(x1,x1i,nx1,pf,o1)

use params_ss_grid
use Numerical_Libraries

implicit none

integer, intent(in)             :: nx1
double precision, intent(in)    :: x1(nx1)
double precision, intent(in)    :: x1i
double precision, intent(in)    :: pf(nx1)
double precision, intent(out)   :: o1

double precision :: s1
double precision :: x1i_min
double precision :: loc1

double precision :: xi
double precision :: xi_left
double precision :: xi_right

double precision :: w_2
double precision :: w_1
double precision :: w1(2)

integer :: m1

o1 = 0

s1      = x1(2) - x1(1)
x1i_min = x1i - x1(1)
loc1    = min(nx1 - 1, max(1,floor(x1i_min/s1)+1))

xi = x1i
xi_left = x1(loc1)
xi_right = x1(loc1 + 1)

w_2 = (xi - xi_left)/(xi_right - xi_left)
w_1 = 1 - w_2

w1(1) = w_1
w1(2) = w_2

do m1 = 0,1
    o1 = o1 + w1(m1+1)*pf(loc1+m1)
end do

return

end subroutine linear_interp_1S

! ************************************************************************
! ---------------------------------------------------------------------
! ---------------------------------------------------------------------
! recursive function : E1IT
! 
! details : computes the time-one expected nominal interest rate at 
! time t
! ---------------------------------------------------------------------
! ---------------------------------------------------------------------

recursive function E1IT(StateToday,IterationToday,rACTUALYest) result (E1IT_result)

use Numerical_Libraries
use params_ss_grid

implicit none

integer, intent(in) :: StateToday, IterationToday
double precision    :: E1IT_result
double precision    :: pR, pG
double precision    :: rACTUALYest
double precision    :: rTod, rACTUALTod

if (IterationToday > 1) then
    call linear_interp_1S(grid_rACTUAL,rACTUALYest,n_rACTUAL,pf_rACTUAL(:,StateToday),rACTUALTod)
    if (StateToday .EQ. 1) then
        pR = pHr
        E1IT_result = (1-pR)*E1IT(1,IterationToday-1,rACTUALTod)
    else
        pR = pLr
        E1IT_result = (1-pR)*E1IT(1,IterationToday-1,rACTUALTod) + pR*E1IT(2,IterationToday-1,rACTUALTod) 
    end if
else
    call linear_interp_1S(grid_rACTUAL,rACTUALYest,n_rACTUAL,pf_r(:,StateToday),rTod) 
    E1IT_result = rTod
end if

end function E1IT

! *********************************************************************
! ---------------------------------------------------------------------
! ---------------------------------------------------------------------
! recursive function : E1PT
! 
! details : computes the time-one expected inflation rate at time t
! ---------------------------------------------------------------------
! ---------------------------------------------------------------------

recursive function E1PT(StateToday,IterationToday,rACTUALYest) result (E1PT_result)

use Numerical_Libraries
use params_ss_grid

implicit none

integer, intent(in) :: StateToday, IterationToday
double precision    :: E1PT_result
double precision    :: pR, pG
double precision    :: rACTUALYest
double precision    :: rACTUALTod, piTod

if (IterationToday > 1) then
    call linear_interp_1S(grid_rACTUAL,rACTUALYest,n_rACTUAL,pf_rACTUAL(:,StateToday),rACTUALTod)
    if (StateToday .EQ. 1) then
        pR = pHr
        E1PT_result = (1-pR)*E1PT(1,IterationToday-1,rACTUALTod)
    else
        pR = pLr
        E1PT_result = (1-pR)*E1PT(1,IterationToday-1,rACTUALTod) + pR*E1PT(2,IterationToday-1,rACTUALTod)
    end if
else
    call linear_interp_1S(grid_rACTUAL,rACTUALYest,n_rACTUAL,pf_pi(:,StateToday),piTod) 
    E1PT_result = piTod
end if

end function E1PT

! *********************************************************************
! ---------------------------------------------------------------------
! ---------------------------------------------------------------------
! recursive function : E1dYT
! 
! details : computes time-one expected output at time t
! ---------------------------------------------------------------------
! ---------------------------------------------------------------------

recursive function E1dYT(StateToday,IterationToday,rACTUALYest) result (E1dYT_result)

use Numerical_Libraries
use params_ss_grid

implicit none

integer, intent(in) :: StateToday, IterationToday
double precision    :: E1dYT_result
double precision    :: pR, pG
double precision    :: rACTUALYest
double precision    :: rACTUALTod, yTod

if (IterationToday > 1) then
    call linear_interp_1S(grid_rACTUAL,rACTUALYest,n_rACTUAL,pf_rACTUAL(:,StateToday),rACTUALTod)
    if (StateToday .EQ. 1) then
        pR = pHr
        E1dYT_result = cBET*(1-pR)*E1dYT(1,IterationToday-1,rACTUALTod)
    else
        pR = pLr
        E1dYT_result = cBET*(1-pR)*E1dYT(1,IterationToday-1,rACTUALTod) + cBET*pR*E1dYT(2,IterationToday-1,rACTUALTod)
    end if    
else
    call linear_interp_1S(grid_rACTUAL,rACTUALYest,n_rACTUAL,pf_y(:,StateToday),yTod) 
    E1dYT_result = yTod
end if

end function E1dYT

! *********************************************************************
! ---------------------------------------------------------------------
! ---------------------------------------------------------------------
! recursive function : E1dGT
! 
! details : computes time-one expected government spending at time t
! ---------------------------------------------------------------------
! ---------------------------------------------------------------------

recursive function E1dGT(StateToday,IterationToday,rACTUALYest) result (E1dGT_result)

use Numerical_Libraries
use params_ss_grid

implicit none

integer, intent(in) :: StateToday, IterationToday
double precision    :: E1dGT_result
double precision    :: pR, pG
double precision    :: rACTUALYest
double precision    :: rACTUALTod, GTod

if (IterationToday > 1) then
    call linear_interp_1S(grid_rACTUAL,rACTUALYest,n_rACTUAL,pf_rACTUAL(:,StateToday),rACTUALTod)
    if (StateToday .EQ. 1) then
        pR = pHr
        E1dGT_result = cBET*(1-pR)*E1dGT(1,IterationToday-1,rACTUALTod)
    else
        pR = pLr
        E1dGT_result = cBET*(1-pR)*E1dGT(1,IterationToday-1,rACTUALTod) + cBET*pR*E1dGT(2,IterationToday-1,rACTUALTod)
    end if    
else
    if (StateToday .EQ. 1) then
        GTod = Lg
    else if (StateToday .EQ. 2) then
        GTod = Hg
    end if
    E1dGT_result = GTod
end if

end function E1dGT